// $Id: Output.c 18 2014-06-16 11:59:04Z dettus $
/*
Copyright (c) 2014, Thomas Dettbarn (dettus@dettus.net)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer. 
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS \"AS IS\" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <ncurses.h>
#include <unistd.h>
#include "Datatypes.h"
#include "Output.h"

void initcolpairs(tHandle* handle)
{
	#define	MKCOLOR(n, FG,BG,ATTR)	\
		handle->colors[n].fg=FG;	\
		handle->colors[n].bg=BG;	\
		handle->colors[n].attr=ATTR;	\
		init_pair(n,FG,BG);		\
		handle->colors[n].combined_attributes=ATTR+COLOR_PAIR(n);

	MKCOLOR(COL_TILE_2	,COLOR_WHITE,COLOR_GREEN,A_BOLD);
	MKCOLOR(COL_TILE_4	,COLOR_WHITE,COLOR_RED  ,A_BOLD);
	MKCOLOR(COL_TILE_8	,COLOR_WHITE,COLOR_CYAN ,A_BOLD);
	MKCOLOR(COL_TILE_16	,COLOR_WHITE,COLOR_MAGENTA,A_BOLD);
	MKCOLOR(COL_TILE_32	,COLOR_WHITE,COLOR_YELLOW,A_BOLD);
	MKCOLOR(COL_TILE_64	,COLOR_WHITE,COLOR_BLACK,A_BOLD);
	MKCOLOR(COL_TILE_128	,COLOR_YELLOW,COLOR_GREEN,A_BOLD);
	MKCOLOR(COL_TILE_256	,COLOR_YELLOW,COLOR_RED,A_BOLD);
	MKCOLOR(COL_TILE_512	,COLOR_YELLOW,COLOR_CYAN,A_BOLD);
	MKCOLOR(COL_TILE_1024	,COLOR_YELLOW,COLOR_MAGENTA,A_BOLD);
	MKCOLOR(COL_TILE_2048	,COLOR_YELLOW,COLOR_BLACK,A_BOLD);
	MKCOLOR(COL_TILE_SPACE	,COLOR_WHITE,COLOR_BLACK,A_BOLD);
	MKCOLOR(COL_FRAME	,COLOR_BLUE,COLOR_BLACK,A_BOLD);
	MKCOLOR(COL_HUD		,COLOR_CYAN,COLOR_BLACK,A_BOLD);
}

void initOutput(tHandle* handle)
{
	if (handle->win==NULL)
	{
		handle->win=initscr();
		start_color();
		noecho();
		nodelay(handle->win,0);
	}
	clear();
	initcolpairs(handle);
}
#define	FIELD_ROWS	4	
#define	FIELD_COLS	6
#define	START_ROW	0
#define	START_COL	0


void drawTile(tHandle* handle,tInt16 tilevalue,int y,int x)
{
	int i,j;
	switch (tilevalue)
	{
		case    0:	wattrset(handle->win,handle->colors[COL_FRAME].combined_attributes);break;
		case    2:	wattrset(handle->win,handle->colors[COL_TILE_2   ].combined_attributes);break;
		case    4:	wattrset(handle->win,handle->colors[COL_TILE_4   ].combined_attributes);break;
		case    8:	wattrset(handle->win,handle->colors[COL_TILE_8   ].combined_attributes);break;
		case   16:	wattrset(handle->win,handle->colors[COL_TILE_16  ].combined_attributes);break;
		case   32:	wattrset(handle->win,handle->colors[COL_TILE_32  ].combined_attributes);break;
		case   64:	wattrset(handle->win,handle->colors[COL_TILE_64  ].combined_attributes);break;
		case  128:	wattrset(handle->win,handle->colors[COL_TILE_128 ].combined_attributes);break;
		case  256:	wattrset(handle->win,handle->colors[COL_TILE_256 ].combined_attributes);break;
		case  512:	wattrset(handle->win,handle->colors[COL_TILE_512 ].combined_attributes);break;
		case 1024:	wattrset(handle->win,handle->colors[COL_TILE_1024].combined_attributes);break;
		case 2048:	wattrset(handle->win,handle->colors[COL_TILE_2048].combined_attributes);break;
	}
	for (i=0;i<FIELD_ROWS;i++)
	{
		wmove(handle->win,y+i,x);
		for (j=0;j<FIELD_COLS;j++)
		{
			waddch(handle->win,' ');
		}
	}		
	if (tilevalue)
	{
		wmove(handle->win,y,x);
		waddch(handle->win,ACS_ULCORNER);
		for (j=0;j<FIELD_COLS-2;j++)
		{
			waddch(handle->win,ACS_HLINE);
		}
		waddch(handle->win,ACS_URCORNER);

		for (i=0;i<FIELD_ROWS-2;i++)
		{
			wmove(handle->win,y+1+i,x);
			waddch(handle->win,ACS_VLINE);
			wmove(handle->win,y+1+i,x+FIELD_COLS-1);
			waddch(handle->win,ACS_VLINE);
		}
		mvwprintw(handle->win,y+FIELD_ROWS/2,x+1,"%4i",tilevalue);	

		wmove(handle->win,y+FIELD_ROWS-1,x);
		waddch(handle->win,ACS_LLCORNER);
		for (j=0;j<FIELD_COLS-2;j++)
		{
			waddch(handle->win,ACS_HLINE);
		}
		waddch(handle->win,ACS_LRCORNER);
	} 
}

void drawFrame(tHandle* handle)
{
	int i,j,k;
	wattrset(handle->win,handle->colors[COL_FRAME].combined_attributes);
	for (i=0;i<TILE_COLS;i++)
	{
		wmove(handle->win,START_ROW,START_COL+(i*(FIELD_COLS+1)));
		if (i) waddch(handle->win,ACS_TTEE);
		else waddch(handle->win,ACS_ULCORNER);
		for (j=0;j<FIELD_COLS;j++)
			waddch(handle->win,ACS_HLINE);
	}
	waddch(handle->win,ACS_URCORNER);

	for (i=0;i<TILE_ROWS;i++)
	{
		if (i)
		{
			wmove(handle->win,START_ROW+(i*(FIELD_ROWS+1)),START_COL);
			waddch(handle->win,ACS_LTEE);
			for (j=0;j<TILE_COLS;j++)
			{
				if (j) waddch(handle->win,ACS_PLUS);
				for (k=0;k<FIELD_COLS;k++) waddch(handle->win,ACS_HLINE);
			}
			waddch(handle->win,ACS_RTEE);

		}
		for (j=0;j<FIELD_ROWS;j++)
		{
			for (k=0;k<TILE_COLS+1;k++)
			{
				wmove(handle->win,START_ROW+(i*(FIELD_ROWS+1))+j+1,START_COL+(k*(FIELD_COLS+1)));
				waddch(handle->win,ACS_VLINE);
			}
		}
	}

	for (i=0;i<TILE_COLS;i++)
	{
		wmove(handle->win,START_ROW+(4*FIELD_ROWS+1)+3,START_COL+(i*(FIELD_COLS+1)));
		if (i) waddch(handle->win,ACS_BTEE);
		else waddch(handle->win,ACS_LLCORNER);
		for (j=0;j<FIELD_COLS;j++)
			waddch(handle->win,ACS_HLINE);
	}
	waddch(handle->win,ACS_LRCORNER);
}

void drawField(tHandle* handle)
{
	int i,j;
	int maxtile;


	maxtile=0;

	for (i=0;i<TILE_ROWS;i++)
	{
		for (j=0;j<TILE_COLS;j++)
		{
			if (maxtile<handle->actField.tile[i][j]) maxtile=handle->actField.tile[i][j];
			drawTile(handle,handle->actField.tile[i][j],START_ROW+(i*(FIELD_ROWS+1))+1,START_COL+1+(j*(FIELD_COLS+1)));
		}
	}
	drawFrame(handle);
	wattrset(handle->win,handle->colors[COL_HUD].combined_attributes);



	mvwprintw(handle->win, 5,45,"W");
	mvwprintw(handle->win, 6,43,"A - D");
	mvwprintw(handle->win, 7,45,"S");
	
	mvwprintw(handle->win,FIELD_ROWS*(TILE_ROWS+1)-TILE_ROWS-1,40,"Score:");
	mvwprintw(handle->win,FIELD_ROWS*(TILE_ROWS+1)-TILE_ROWS,40,"%6i",handle->score);
	mvwprintw(handle->win,FIELD_ROWS*(TILE_ROWS+1)-TILE_ROWS-1,48,"Best:");
	mvwprintw(handle->win,FIELD_ROWS*(TILE_ROWS+1)-TILE_ROWS,47,"%6i",handle->best);
	mvwprintw(handle->win,FIELD_ROWS*(TILE_ROWS+1)-TILE_ROWS-1,64+TILE_COLS-10,"Best tile:");
	drawTile(handle,maxtile,FIELD_ROWS*(TILE_ROWS+1)-TILE_ROWS,62);
	
}

int winner(tHandle* handle)
{

	int i;
	tInt16	ch;
	tInt16	colors[8]={COLOR_BLACK,COLOR_WHITE,COLOR_RED,COLOR_GREEN,COLOR_BLUE,COLOR_YELLOW,COLOR_CYAN,COLOR_MAGENTA};

	nodelay(handle->win,1);
	ch=~ERR;
	while (ch!=ERR) ch=getch();
	
	ch=ERR;
	i=0;
	init_pair(17,COLOR_WHITE,COLOR_BLACK);
	wattrset(handle->win,COLOR_PAIR(17)+A_BOLD);
	mvwprintw(handle->win,12,50,"Another game? Y/N");
	while (ch!='y' && ch!='Y' && ch!='n' && ch!='N')
	{
		usleep(1000);
		ch=getch();
		init_pair(16,colors[i],COLOR_BLACK);
		wattrset(handle->win,COLOR_PAIR(16)+A_BOLD);
		mvwprintw(handle->win, 5,10,"_______________      _____   ______  ");
		mvwprintw(handle->win, 6,10,"\\_____  \\   _  \\    /  |  | /  __  \\ ");
		mvwprintw(handle->win, 7,10," /  ____/  /_\\  \\  /   |  |_>      < ");
		mvwprintw(handle->win, 8,10,"/       \\  \\_/   \\/    ^   /   --   \\");
		mvwprintw(handle->win, 9,10,"\\_______ \\_____  /\\____   |\\______  /");
		mvwprintw(handle->win,10,10,"        \\/     \\/      |__|       \\/ ");

		i++;
		if (i==8) i=0;

	}
	if (ch=='n' || ch=='N') return 1;
	return 0;
}
int loser(tHandle* handle)
{

	int i;
	tInt16	ch;

	nodelay(handle->win,1);

	ch=~ERR;
	while (ch!=ERR) ch=getch();
	ch=ERR;
	i=0;
	init_pair(17,COLOR_WHITE,COLOR_BLACK);
	wattrset(handle->win,COLOR_PAIR(17)+A_BOLD);
	mvwprintw(handle->win,12,50,"Another game? Y/N");
	while (ch!='y' && ch!='Y' && ch!='n' && ch!='N')
	{
		usleep(100000);
		ch=getch();
		switch(i)
		{
			case 0: init_pair(16,COLOR_BLACK,COLOR_BLACK); wattrset(handle->win,COLOR_PAIR(16)); break;
			case 1: init_pair(16,COLOR_BLACK,COLOR_BLACK); wattrset(handle->win,COLOR_PAIR(16)+A_BOLD); break;
			case 2: init_pair(16,COLOR_WHITE,COLOR_BLACK); wattrset(handle->win,COLOR_PAIR(16)); break;
			case 3: init_pair(16,COLOR_WHITE,COLOR_BLACK); wattrset(handle->win,COLOR_PAIR(16)+A_BOLD); break;
			case 4: init_pair(16,COLOR_WHITE,COLOR_BLACK); wattrset(handle->win,COLOR_PAIR(16)+A_BOLD); break;
			case 5: init_pair(16,COLOR_WHITE,COLOR_BLACK); wattrset(handle->win,COLOR_PAIR(16)); break;
			case 6: init_pair(16,COLOR_BLACK,COLOR_BLACK); wattrset(handle->win,COLOR_PAIR(16)+A_BOLD); break;
			case 7: init_pair(16,COLOR_BLACK,COLOR_BLACK); wattrset(handle->win,COLOR_PAIR(16)); break;

		}
		mvwprintw(handle->win, 5,10,"  ________                        ________");
		mvwprintw(handle->win, 6,10," /  _____/_____    _____   ____   \\_____  \\___  __ ___________ ");
		mvwprintw(handle->win, 7,10,"/   \\  ___\\__  \\  /     \\_/ __ \\   /   |   \\  \\/ // __ \\_  __ \\");
		mvwprintw(handle->win, 8,10,"\\    \\_\\  \\/ __ \\|  Y Y  \\  ___/  /    |    \\   /\\  ___/|  | \\/");
		mvwprintw(handle->win, 9,10," \\______  (____  /__|_|  /\\___  > \\_______  /\\_/  \\___  >__|   ");
		mvwprintw(handle->win,10,10,"        \\/     \\/      \\/     \\/          \\/          \\/       ");

		i++;
		if (i==8) i=0;

	}
	if (ch=='n' || ch=='N') return 1;
	return 0;
}


void endOutput(tHandle* handle)
{
	endwin();
}

void transition(tHandle* handle,int value, int from_row, int to_row, int from_col, int to_col)
{
	int act_x;
	int act_y;
	int to_x;
	int to_y;

	act_x=START_COL+(from_col*(FIELD_COLS+1))+1;
	to_x=START_COL+(to_col*(FIELD_COLS+1))+1;
	act_y=START_ROW+(from_row*(FIELD_ROWS+1))+1;
	to_y=START_ROW+(to_row*(FIELD_ROWS+1))+1;

	
	while (act_x!=to_x || act_y!=to_y)
	{
		drawTile(handle,0,act_y,act_x);
		if (act_y>to_y) act_y--;
		if (act_y<to_y) act_y++;
		if (act_x>to_x) act_x--;
		if (act_x<to_x) act_x++;
		drawTile(handle,value,act_y,act_x);
		drawFrame(handle);
		refresh();
		usleep(2400);
	}
}

